/*
 * ==========================================================================
 *
 *       Filename:  main.c
 *
 *    Description:  main file for "my_complex plugin"
 *
 *        Version:  1.0
 *        Created:  08/08/2021 05:16:29 PM
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Haoyang Liu (@liuhy), liuhaoyang@pku.edu.cn
 *   Organization:  BICMR, PKU
 *
 * ==========================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bex/bex.h"

#define PLUGIN_NAME "my_complex"

// 定义 bit 兼容的 complex 类型
// 使用 bxGetComplexDoubles 返回的类型可以强制转化为 double[2]
typedef double complex_baltam[2];

/**
 * @brief main.c 示例一个 BALTAM 插件（工具箱）的开发方法，
 * 该插件使用 C 语言以及 C 编译器，其功能是设计另外一种复数数据结构，
 * 将复数数组的实部和虚部分开存储，而不是存成 double[2] 的形式。
 *
 * 实现的函数：
 * A = bc_zeros(n)
 * A = bc_create_from_baltam(B)
 * B = bc_cvt_to_baltam(A)
 * bc_show(A)
 * N = bc_size(A)
 *
 */
void bc_zeros(int, bxArray*[], int, const bxArray*[]);
void bc_create_from_baltam(int, bxArray*[], int, const bxArray*[]);
void bc_cvt_to_baltam(int, bxArray*[], int, const bxArray*[]);
void bc_size(int, bxArray*[], int, const bxArray*[]);
void bc_show(int, bxArray*[], int, const bxArray*[]);

/**
 * @brief my_complex 自定义复数数组，
 * 它的实部和虚部是分开放置的。
 */
typedef struct {
    double *real;
    double *imag;
    baSize N;
} my_complex;

// 定义复制
void *my_complex_dup(const void *x){
    const my_complex *X = (const my_complex*)x;
    my_complex *ret = (my_complex*)malloc(sizeof(my_complex));
    ret->N = X->N;

    // 创建
    ret->real = (double*)malloc(ret->N * sizeof(double));
    ret->imag = (double*)malloc(ret->N * sizeof(double));

    // 复制
    memcpy(ret->real, X->real, X->N * sizeof(double));
    memcpy(ret->imag, X->imag, X->N * sizeof(double));

    return ret;
}

// 定义析构
void my_complex_destroy(void *x){
    my_complex *X = (my_complex*)x;
    free(X->real); X->real = NULL;
    free(X->imag); X->imag = NULL;
    // 这里的 free(X) 是必要的
    // BALTAM 无法得知原始的 x 是如何分配内存的，
    // 因此交给开发者处理
    free(X);
}

/**
 * @brief 初始化 sid，记录 cstruct 信息
 */
int my_sid = 0;

/**
 * @brief bxPluginInit 的定义（可选）
 *
 * bxPluginInit(args...) 会在调用 load_plugin(name, args...) 时调用，目的是
 * 进行一些初始化工作。
 *
 * 这里的初始化工作是向内核添加 my_complex 这个结构体
 */

int bxPluginInit(int nrhs, const bxArray *prhs[]){
    my_sid = bxRegisterCStruct(PLUGIN_NAME, my_complex_dup, my_complex_destroy);
    return 0;
}

/**
 * @brief bxPluginFini 的定义（可选）
 *
 * bxPluginFini() 会在调用 unload_plugin() 时调用。目的是进行一些析构的工作。
 *
 * 这里因为无需特别的析构，所以定义为空函数。
 */

int bxPluginFini(){ return 0; }

/**
 * @brief bxPluginFunctions 的定义（必要）
 *
 * bxPluginFunctions 返回一个 bexfun_info_t * 的指针，指向函数列表的第一个元素。
 * 列表中最后一个元素的 name 标签必须为空字符串常量。
 * 函数列表可以以静态变量的方式定义在 .cpp 或 .c 中。
 *
 */
static bexfun_info_t flist[] = {
    {"bc_zeros", bc_zeros, NULL},
    {"bc_create_from_baltam", bc_create_from_baltam, NULL},
    {"bc_cvt_to_baltam", bc_cvt_to_baltam, NULL},
    {"bc_size", bc_size, NULL},
    {"bc_show", bc_show, NULL},
    {"", NULL, NULL},
};

bexfun_info_t * bxPluginFunctions(){
    return flist;
}

/**
 * @brief 以下是各种函数的实际定义。
 */

BALTAM_PLUGIN_FCN(bc_zeros){
    if (nlhs > 1 || nrhs != 1)
        return;

    baSize N = (baSize)(*bxGetDoubles(prhs[0]));
    if (N <= 0)
        return;
    my_complex *mc = (my_complex*)malloc(sizeof(my_complex));
    mc->N = N;
    mc->real = (double*)calloc(N, sizeof(double));
    mc->imag = (double*)calloc(N, sizeof(double));

    plhs[0] = bxCreateCStruct(my_sid, mc);
}

BALTAM_PLUGIN_FCN(bc_show){
    if (nrhs != 1 || nlhs != 0)
        return;

    my_complex *ptr = (my_complex*)bxGetCStruct(my_sid, prhs[0]);
    
    if (ptr == NULL)
        return;

    for (baIndex i = 0; i < ptr->N; ++i){
        printf("[%ld]    (%6.2f, %6.2f)\n", i, ptr->real[i], ptr->imag[i]);
    }
}

BALTAM_PLUGIN_FCN(bc_create_from_baltam){
    if (nlhs > 1 || nrhs != 1)
        return;

    baSize M = bxGetM(prhs[0]);
    baSize N = bxGetN(prhs[0]);

    if (M != 1 || N <= 0)
        return;

    complex_baltam *dat = (complex_baltam*)bxGetComplexDoubles(prhs[0]);
    my_complex *mc = (my_complex*)malloc(sizeof(my_complex));
    mc->N = N;
    mc->real = (double*)malloc(N * sizeof(double));
    mc->imag = (double*)malloc(N * sizeof(double));

    for (baIndex i = 0; i < N; ++i){
        mc->real[i] = dat[i][0];
        mc->imag[i] = dat[i][1];
    }

    plhs[0] = bxCreateCStruct(my_sid, mc);

}

BALTAM_PLUGIN_FCN(bc_cvt_to_baltam){
    if (nlhs > 1 || nrhs != 1)
        return;

    my_complex *mc = (my_complex*)bxGetCStruct(my_sid, prhs[0]);

    if (mc == NULL)
        return;

    plhs[0] = bxCreateDoubleMatrix(1, mc->N, bxCOMPLEX);
    complex_baltam *dat = (complex_baltam*)bxGetComplexDoubles(plhs[0]);

    for (baIndex i = 0; i < mc->N; ++i){
        dat[i][0] = mc->real[i];
        dat[i][1] = mc->imag[i];
    }

}

BALTAM_PLUGIN_FCN(bc_size){
    if (nlhs > 1 || nrhs != 1)
        return;

    my_complex *mc = (my_complex*)bxGetCStruct(my_sid, prhs[0]);

    if (mc == NULL)
        return;

    plhs[0] = bxCreateDoubleMatrix(1, 1, bxREAL);
    double *pr = bxGetDoubles(plhs[0]);

    *pr = (double)mc->N;
}
