/*
 * ==========================================================================
 *
 *       Filename:  main.cpp
 *
 *    Description:  main file for "point plugin"
 *
 *        Version:  1.0
 *        Created:  09/28/2021 04:04:00 PM
 *       Revision:  none
 *       Compiler:  g++
 *
 *         Author:  Haoyang Liu (@liuhy), liuhaoyang@pku.edu.cn
 *   Organization:  BICMR, PKU
 *
 * ==========================================================================
 */

#include <iostream>
#include <sstream>
#include <vector>
#include "bex/bex.hpp"

#define PLUGIN_NAME "point"

using namespace baltam;

/**
 * @brief main.cpp 示例一个 BALTAM 插件（工具箱）的开发方法，其功能是
 * 演示如何重载 BALTAM 中的算子。
 *
 * 实现的函数：
 * p = create_point(x, y);
 * show_point(p);
 * p3 = add_point(p1, p2);
 *
 */
void create_point(int, bxArray*[], int, const bxArray*[]);
void show_point(int, bxArray*[], int, const bxArray*[]);
void add_point(int, bxArray*[], int, const bxArray*[]);

/**
 * @brief 各个函数的帮助文字（可选）
 */
static const char * create_help = R"(
create_point 创建一个 point 类型对象

    p = create_point(x, y) 创建一个 point 对象，坐标分别为 x 和 y。
    要求 x 和 y 必须为 double 类型。
)";

static const char * show_help = R"(
show_point 在命令行中显示一个 point 的简略信息。

    bv_show(p) 在命令行中 (x, y) 的方式来显示 p。

)";

/**
 * @brief Point 自定义类。
 *
 * 作为一种自定义数据类型，Point 必须继承自 batlam::extern_obj_base 类。
 * 并定义 ID 这一静态变量（会被内核使用）。
 */

struct Point : public extern_obj_base {
    double x_;
    double y_;
    BALTAM_LOCAL static int ID;

    extern_obj_base *dup() const override;
    Point();
    Point(double, double);
    Point(const Point&) = default;
    ~Point() override;
    // 可选：转化为字符串
    std::string to_string() const override;
    // 可选：类型名字
    std::string classname() const override;
};

int Point::ID = 0;

Point::Point() : x_(0), y_(0) {}
Point::Point(double x, double y) : x_(x), y_(y) {}

extern_obj_base *Point::dup() const {
    return new Point(*this);
}

std::string Point::to_string() const {
    std::stringstream ss;
    ss << '(' << x_ << ", " << y_ << ')';
    return ss.str();
}

std::string Point::classname() const {
    return "Point";
}

Point::~Point() {}
/**
 * @brief bxPluginInit 的定义（可选）
 *
 * bxPluginInit(args...) 会在调用 load_plugin(name, args...) 时调用，目的是
 * 进行一些初始化工作。
 *
 * 这里的初始化工作是向内核添加 Point 这个类
 * 并重载 + (bxADD_OP) 算子
 */

int bxPluginInit(int, const bxArray*[]){
    int sid = bxAddCXXClass<Point>(PLUGIN_NAME);
    bxRegisterBinaryOperator(PLUGIN_NAME, "+", sid, sid, add_point);
    return 0;
}

/**
 * @brief bxPluginFini 的定义（可选）
 *
 * bxPluginFini() 会在调用 unload_plugin() 时调用。目的是进行一些析构的工作。
 *
 * 这里因为无需特别的析构，所以定义为空函数。
 */

int bxPluginFini(){ return 0; }

/**
 * @brief bxPluginFunctions 的定义（必要）
 *
 * bxPluginFunctions 返回一个 bexfun_info_t * 的指针，指向函数列表的第一个元素。
 * 列表中最后一个元素的 name 标签必须为空字符串常量。
 * 函数列表可以以静态变量的方式定义在 .cpp 或 .c 中。
 *
 */
static bexfun_info_t flist[] = {
    {"create_point", create_point, create_help},
    {"show_point", show_point, show_help},
    {"", nullptr, nullptr},
};

bexfun_info_t * bxPluginFunctions(){
    return flist;
}

/**
 * @brief 以下是各种函数的实际定义。
 */

void create_point(int nlhs, bxArray *plhs[], int nrhs, const bxArray *prhs[]){
    if (nlhs > 1 && nrhs != 2)
        return;

    if (!bxIsDouble(prhs[0]) || !bxIsDouble(prhs[1]))
        return;

    Point * ret = bxNewCXXObject<Point>(
            *bxGetDoubles(prhs[0]),
            *bxGetDoubles(prhs[1]));
    plhs[0] = bxCreateExtObj(ret);
}

void show_point(int, bxArray *[], int nrhs, const bxArray *prhs[]){
    if (nrhs != 1){
        return;
    }

    Point *bv = bxGetExtObj<Point>(prhs[0]);
    std::cout << bv->to_string() << std::endl;
}


void add_point(int nlhs, bxArray *plhs[], int nrhs, const bxArray *prhs[]){
    if (nlhs > 1 && nrhs != 2)
        return;
    
    Point *p1 = bxGetExtObj<Point>(prhs[0]);
    Point *p2 = bxGetExtObj<Point>(prhs[1]);

    Point *p3 = bxNewCXXObject<Point>(p1->x_ + p2->x_, p1->y_ + p2->y_);
    plhs[0] = bxCreateExtObj(p3);
}
