/*
 * ==========================================================================
 *
 *       Filename:  main.cpp
 *
 *    Description:  main file for "vector plugin"
 *
 *        Version:  1.0
 *        Created:  07/31/2021 09:04:43 PM
 *       Revision:  none
 *       Compiler:  g++
 *
 *         Author:  Haoyang Liu (@liuhy), liuhaoyang@pku.edu.cn
 *   Organization:  BICMR, PKU
 *
 * ==========================================================================
 */

#include <iostream>
#include <sstream>
#include <vector>
#include "bex/bex.hpp"
#ifdef BV_USE_DYN_LOADER
#include "bex/bex.import.hpp"
#endif


#define PLUGIN_NAME "vector"

using namespace baltam;

/**
 * @brief main.cpp 示例一个 BALTAM 插件（工具箱）的开发方法，其功能是
 * 能够在 BALTAM 命令行里使用类似 std::vector 这样的数据结构。
 *
 * 实现的函数：
 * bV = bv_create(...)
 * N = bv_size(bV)
 * obj = bv_at(bV);
 * bv_show(bV);
 *
 */
void bv_create(int, bxArray*[], int, const bxArray*[]);
void bv_size(int, bxArray*[], int, const bxArray*[]);
void bv_at(int, bxArray*[], int, const bxArray*[]);
void bv_show(int, bxArray*[], int, const bxArray*[]);

/**
 * @brief 各个函数的帮助文字（可选）
 */
static const char * create_help = R"(
bv_create 创建一个 vector

    V = bv_create(...) 创建一个 vector，元素为所有输入参数。
    V 中的元素是输入参数的复制。
)";

static const char * size_help = R"(
bv_size 获取一个 vector 的元素个数

    N = bv_size(V) 获取 V 的元素数量，返回一个整数。
)";

static const char * show_help = R"(
bv_show 在命令行中显示一个 vector 的简略信息。

    bv_show(V) 在命令行中以位置 -- 简略信息的方式来显示 V。

)";

static const char * at_help = R"(
bv_at 获取一个 vector 指定位置的元素。

    elem = bv_at(V, ind) 获取 V 中索引为 ind 的元素的拷贝。索引 ind
    从 0 开始。

    当 ind 发生越界时，bv_at 会报错。
)";

#ifdef BV_USE_DYN_LOADER
/**
 * @brief 定义 bex.import.hpp 中的所有函数变量
 */
BALTAM_BEX_DEFINE_FCN_VARS
/**
 * @brief bxPluginInitLib 的定义（可选）
 *
 * bxPluginInitLib(hdl) 会在调用 load_plugin(name, args...) 时调用，
 * 且先于 bxPluginInit 调用。
 *
 * 主要目的是在动态载入的模式下，获取 libbex 的信息并读取其中的函数。
 */
int bxPluginInitLib(void * hdl){
    bex::bex_handle = hdl;
    BALTAM_BEX_IMPORT_FCN_FROM(hdl)
    return 0;
}
#endif

/**
 * @brief bVector 自定义 vector 类，实际是对 std::vector 的包装。
 *
 * 作为一种自定义数据类型，bVector 必须继承自 batlam::extern_obj_base 类。
 * 并定义 ID 这一静态变量（会被内核使用）。
 */

struct bVector : public extern_obj_base {
    std::vector<bxArray*> data;
    BALTAM_LOCAL static int ID;
    extern_obj_base *dup() const override;
    ~bVector() override;
    // 可选：转化为字符串的实现
    std::string to_string() const override;
    // 可选：自定义类型名字
    std::string classname() const override;
};

int bVector::ID = 0;

extern_obj_base *bVector::dup() const {
    bVector *ret = new bVector(*this);
    for (size_t i = 0; i < this->data.size(); ++i){
#ifdef BV_USE_DYN_LOADER
        ret->data[i] = bex::__bxDuplicateArray(data[i]);
#else
        ret->data[i] = bxDuplicateArray(data[i]);
#endif
    }
    return ret;
}

bVector::~bVector(){
    for (auto &i : data){
#ifdef BV_USE_DYN_LOADER
        bex::__bxDestroyArray(i);
#else
        bxDestroyArray(i);
#endif
    }
}

std::string bVector::to_string() const {
    std::stringstream ss;
    for (size_t i = 0; i < data.size(); ++i){
#ifdef BV_USE_DYN_LOADER
        ss << '[' << i << "]    " << bex::__bxTypeCStr(data[i]) << std::endl;
#else
        ss << '[' << i << "]    " << bxTypeString(data[i]) << std::endl;
#endif
    }
    return ss.str();
}

std::string bVector::classname() const {
    return "bVector";
}

/**
 * @brief bxPluginInit 的定义（可选）
 *
 * bxPluginInit(args...) 会在调用 load_plugin(name, args...) 时调用，目的是
 * 进行一些初始化工作。
 *
 * 这里的初始化工作是向内核添加 bVector 这个类
 */

int bxPluginInit(int, const bxArray*[]){
#ifdef BV_USE_DYN_LOADER
    bxAddCXXClass<bVector>(PLUGIN_NAME, bex::__bxAddCXXClass_impl);
#else
    bxAddCXXClass<bVector>(PLUGIN_NAME);
#endif
    return 0;
}

/**
 * @brief bxPluginFini 的定义（可选）
 *
 * bxPluginFini() 会在调用 unload_plugin() 时调用。目的是进行一些析构的工作。
 *
 * 这里因为无需特别的析构，所以定义为空函数。
 */

int bxPluginFini(){ return 0; }

/**
 * @brief bxPluginFunctions 的定义（必要）
 *
 * bxPluginFunctions 返回一个 bexfun_info_t * 的指针，指向函数列表的第一个元素。
 * 列表中最后一个元素的 name 标签必须为空字符串常量。
 * 函数列表可以以静态变量的方式定义在 .cpp 或 .c 中。
 *
 */
static bexfun_info_t flist[] = {
    {"bv_create", bv_create, create_help},
    {"bv_size", bv_size, size_help},
    {"bv_at", bv_at, at_help},
    {"bv_show", bv_show, show_help},
    {"", nullptr, nullptr},
};

bexfun_info_t * bxPluginFunctions(){
    return flist;
}

/**
 * @brief 以下是各种函数的实际定义。
 */

void bv_create(int nlhs, bxArray *plhs[], int nrhs, const bxArray *prhs[]){
    if (nlhs > 1)
        return;

    bVector * ret = bxNewCXXObject<bVector>();
#ifdef BV_USE_DYN_LOADER
    for (int i = 0; i < nrhs; ++i){
        bxArray * elem = bex::__bxDuplicateArray(prhs[i]);
        ret->data.push_back(elem);
    }
    plhs[0] = bex::__bxCreateExtObj_v(ret);
#else
    for (int i = 0; i < nrhs; ++i){
        bxArray * elem = bxDuplicateArray(prhs[i]);
        ret->data.push_back(elem);
    }
    plhs[0] = bxCreateExtObj(ret);
#endif
}

void bv_size(int nlhs, bxArray *plhs[], int nrhs, const bxArray *prhs[]){
    if (nlhs > 1 || nrhs != 1){
        return;
    }

#ifdef BV_USE_DYN_LOADER
    bVector *bv = bxGetExtObj<bVector>(prhs[0], bex::__bxGetExtObj_impl);
    if (bv == nullptr)
        return;

    plhs[0] = bex::__bxCreateDoubleMatrix(1, 1, bxREAL);
    double *ptr = bex::__bxGetDoubles(plhs[0]);
#else
    bVector *bv = bxGetExtObj<bVector>(prhs[0]);
    if (bv == nullptr)
        return;

    plhs[0] = bxCreateDoubleMatrix(1, 1, bxREAL);
    double *ptr = bxGetDoubles(plhs[0]);
#endif

    *ptr = static_cast<double>(bv->data.size());
}

void bv_show(int, bxArray *[], int nrhs, const bxArray *prhs[]){
    if (nrhs > 1){
        return;
    }

#ifdef BV_USE_DYN_LOADER
    bVector *bv = bxGetExtObj<bVector>(prhs[0], bex::__bxGetExtObj_impl);

    for (size_t i = 0; i < bv->data.size(); ++i){
        std::cout << '[' << i << "]    " << bex::__bxTypeCStr(bv->data[i]) << std::endl;
#else
    bVector *bv = bxGetExtObj<bVector>(prhs[0]);

    for (size_t i = 0; i < bv->data.size(); ++i){
        std::cout << '[' << i << "]    " << bxTypeString(bv->data[i]) << std::endl;
#endif
    }
}

void bv_at(int nlhs, bxArray *plhs[], int nrhs, const bxArray *prhs[]){
#ifdef BV_USE_DYN_LOADER
    if (nlhs > 1 || nrhs != 2){
        bex::__bxErrMsgTxt("用法：A = bv_at(V, ind)");
        return;
    }

    bVector *bv = bxGetExtObj<bVector>(prhs[0], bex::__bxGetExtObj_impl);

    if (bv == nullptr){
        bex::__bxErrMsgTxt("参数 1 不是合法类型");
        return;
    }

    int err;
    baInt id = bex::__bxAsInt(prhs[1], &err);
    if (err != 0){
        bex::__bxErrMsgTxt("参数 2 不是合法类型");
        return;
    }

    // 检查越界
    if (id >= static_cast<baInt>(bv->data.size()) || id < 0){
        bex::__bxErrMsgTxt("索引越界");
        return;
    }

    plhs[0] = bex::__bxDuplicateArray(bv->data[id]);
#else
    if (nlhs > 1 || nrhs != 2){
        bxErrMsgTxt("用法：A = bv_at(V, ind)");
        return;
    }

    bVector *bv = bxGetExtObj<bVector>(prhs[0]);

    if (bv == nullptr){
        bxErrMsgTxt("参数 1 不是合法类型");
        return;
    }

    int err;
    baInt id = bxAsInt(prhs[1], &err);
    if (err != 0){
        bxErrMsgTxt("参数 2 不是合法类型");
        return;
    }

    // 检查越界
    if (id >= static_cast<baInt>(bv->data.size()) || id < 0){
        bxErrMsgTxt("索引越界");
        return;
    }

    plhs[0] = bxDuplicateArray(bv->data[id]);
#endif

}

#if defined(BV_USE_DYN_LOADER) && defined(BV_BUILD_EXE)
int main(){
    std::vector<const bxArray*> in_args;
    std::vector<bxArray*> out_args;
    // 手动调用 init 函数
#if defined(__WIN32__) || defined(__CYGWIN__) || defined(__MINGW32__)
    void * handle = (void*)LoadLibrary("libbex.dll");
#else
    void * handle = dlopen("libbex.so", RTLD_NOW);
#endif
    bxPluginInitLib(handle);
    // 这里 addCXXClass 其实不会被执行，因为已经绕过了插件管理器
    bxPluginInit(0, nullptr);
    std::cout << "bex 库已载入" << std::endl;

    // 从内核创建数据
    bxArray *A = bex::__bxCreateNumericMatrix(1, 1, bxINT_CLASS, bxREAL);
    bxArray *B = bex::__bxCreateNumericMatrix(1, 1, bxDOUBLE_CLASS, bxREAL);
    bxArray *C = bex::__bxCreateString("test");

    // 修改数据
    int *pA = bex::__bxGetInt32s(A);
    *pA = 42;
    double *pB = bex::__bxGetDoubles(B);
    *pB = 3.14;

    // 调用 bv_create
    in_args.push_back(A);
    in_args.push_back(B);
    in_args.push_back(C);
    out_args.push_back(nullptr);
    bv_create(out_args.size(), out_args.data(), in_args.size(), in_args.data());

    bxArray *V = out_args[0];

    // 调用 bv_size
    in_args.clear();
    out_args.clear();
    in_args.push_back(V);
    out_args.push_back(nullptr);
    bv_size(out_args.size(), out_args.data(), in_args.size(), in_args.data());

    bxArray *N = out_args[0];

    std::cout << "数组的大小是：" << static_cast<int>(*bex::__bxGetDoubles(N)) << std::endl;

    // 调用 bv_show
    in_args.clear();
    out_args.clear();
    in_args.push_back(V);
    bv_show(out_args.size(), out_args.data(), in_args.size(), in_args.data());

    // 调用 bv_at
    in_args.clear();
    out_args.clear();
    bxArray *ii = bex::__bxCreateNumericMatrix(1, 1, bxDOUBLE_CLASS, bxREAL);
    double *pii = bex::__bxGetDoubles(ii);
    *pii = 2;
    in_args.push_back(V);
    in_args.push_back(ii);
    out_args.push_back(nullptr);
    bv_at(out_args.size(), out_args.data(), in_args.size(), in_args.data());

    bxArray *ch = out_args[0];
    char *t = bex::__bxGetChars(ch);
    std::cout << t[0] << t[1] << t[2] << t[3] << std::endl;

    // 释放内存
    bex::__bxDestroyArray(A);
    bex::__bxDestroyArray(B);
    bex::__bxDestroyArray(C);
    bex::__bxDestroyArray(V);
    bex::__bxDestroyArray(N);
    bex::__bxDestroyArray(ii);
    bex::__bxDestroyArray(ch);

#if defined(__WIN32__) || defined(__CYGWIN__) || defined(__MINGW32__)
    FreeLibrary((HMODULE)handle);
#else
    dlclose(handle);
#endif

    return 0;
}
#endif
