% 对已知方差的单个总体均值进行Z检验, 调用格式为
% [h,sig,ci,zval] = ztest(x,mu0,sigma,alpha,tail)
% 进行显著性水平为alpha的Z假设检验, 用来检验标准差为sigma时的正态分布样本的均值与mu0的关系，
% 可通过指定tail的值来选择备择假设的类型：tail = 0 或 'both'，检验假设“x 的均值等于mu0”为默认设置，双侧检验；  tail = 1或 'right'，检验假设“x 的均值大于mu0”，右侧检验； tail =-1或 'left'，检验假设“x 的均值小于mu0”，左侧检验； 
%输入参数 tail: 缺省值为 0 (对应的假设\mu = \mu0), -1 (左侧检验，H1: \mu<\mu0), 1(右侧检验, H1: \mu>\mu0)， 
% 输入参数 alpha: 缺省值为 0.05.  
%返回值 h: 为一个布尔值，h=1 表示可以拒绝假设，h=0 表示不可以拒绝假设，
%返回值 sig: 为假设成立的概率，
%返回值 ci: 样本均值mu的 1-alpha 置信区间, 
%返回值 zval: z统计量的值.
%例
% X= 0.497,  0.506,  0.518,  0.524,  0.498,  0.511,  0.52,  0.515,  0.512
%判读 \mu =0.5 ? 为此提出假设：
%原假设：\mu = 0.5  
%备择假设： \mu \neq 0.5
%X=[0.497,0.506,0.518,0.524,0.498,0.511,0.52,0.515,0.512];
%[h,sig,ci,zval]=ztest(X,0.5,0.015,0.05,0)
%结果显示为
%h =
%     1
%sig =
%    0.0248       %样本观察值的概率
%ci =
%    0.5014    0.5210        %置信区间，均值0.5在此区间之外
%zval =
%    2.2444       %统计量的值
%结果表明：h=1，说明在水平下，可拒绝原假设，即认为包装机工作不正常。

function [h,sig,ci,zval] = ztest(x,mu0,sigma,alpha,tail)
	if nargin < 1 || numel(x) == 0 
		error("ztest 至少需要一个参数, 而且参数的长度要大于0");
	end

	if nargin < 2
		mu0 = 0;
	end

	if(nargin < 3)
		sigma = 1;
	end

	if(nargin < 4) 
		alpha = 0.05;
	end
	if(nargin < 5)
		tail = 0;
	end

	x = x(:); % 先把变量转成一个列向量
	n = length(x); 
	mu = mean(x);

	zval =  (mu-mu0)/(sigma/sqrt(n));

	if strcmpi(tail, 'both') || tail == 0 
		ia = norminv(1-alpha/2);
		zi = [-ia, ia]; 
		ci = mu+zi*sigma/sqrt(n); 
		if zval <= zi(1) || zval >= zi(2)
			h = true;
		else
			h = false;
		end
		%假设 zval \in [zi(1),zi(2)], 
		% 现在看到 zval, 那么计算 z \in (-inf,-|zval|) \cup (|zval|,inf) 的概率
		sig = 2*(1-normcdf( abs(zval) ));
		return;
	end

	if strcmpi(tail, 'right') || tail == 1 
		ia = norminv(1-alpha);
		zi = [-inf, ia]; 
		ci = mu+zi*sigma/sqrt(n); 
		if zval > zi(2)
			h = true;
		else
			h = false;
		end
		%假设 zval \in (zi(1),inf), 
		% 现在看到 zval, 那么计算 z \in (-inf, zval) 的概率
		sig = 1 - normcdf(zval);
		return;
	end

	if strcmpi(tail, 'left') || tail == -1 
		ia = norminv(alpha);
		zi = [ia,inf];
		ci = mu+zi*sigma/sqrt(n);
		if zval < zi(1)
			h = true;
		else
			h = false;
		end
		%假设 zval \in (-inf, zi(2)), 
		% 现在看到 zval, 那么计算 z \in [zval,inf) 的概率
		sig = normcdf(zval);
		return;
	end
end
