function [r,a] = roots(c)
%ROOTS  求多项式的所有根.
%[r,a] = roots(c) 以列向量的形式返回 c 表示的多项式
%  c(1) x^(n) + c(2)x^n-1 + ... + c(n+1)  = 0 
% 的根r 和 多项式的友矩阵a.
%输入 c 是一个包含 n 次多项式系数的向量，以 x^n 系数开头. 0 系数表示方程中不存在的中间幂.
%例如：c = [3,2,-2] 表示多项式 3x^2+2x−2.
%
%   注意:  c 中真正起作用的分量是从第一个非0的分量开始的，前面的零被剔除了。第一个系数虽然不是0，
%         但是如果存在某个系数被第一个系数除之后得到的结果是上溢的，那么，第一个
%         位置的系数也要被剔除调。这个过程一直继续到上溢不发生为止。

% ROOTS 通过求多项式的友矩阵的特征值来求多项式的根.

if nargin ~= 1 
	error('非法的使用roots函数，正确的用法是 roots(c)');
end

if ~exist('c','var')
	error('roots 的输入参数没有被赋值');
end

if ~isempty(c) && ~isvector(c)
    error('roots的输入参数不是向量');
end

if ~all(isfinite(c))
    error('roots的输入参数不全是有限数');
end


c = c(:).'; % c(:)得到的是列向量，然后做了转置，得到了一个行向量
n = size(c,2);  % size(c,2) 得到是矩阵的列数，而c是一个行向量，这得到就是向量的长.　 

inz = find(c);  
if isempty(inz)
    % c 的所有的系数都是0
    r = zeros(0,1,class(c)); % 返回0x1 类型和c相同的矩阵　  
    return;
end

Matc = zeros(size(c),class(c)); 
fullPrototype = full(Matc);

% 把系数c中第一个不为零前面的零系数都剔除
nnz = length(inz);
c = c(inz(1):inz(nnz));
r = zeros(n-inz(nnz),1,class(fullPrototype));
% 如果第一个不为零的系数作为除数去除后面的系数会发生上溢出，则把这个系数也剔除，
% 直到第一个不为零的系数足够大到上溢不再发生为止。但是此时因为上溢抛弃的系数
% 的个数会在返回的结果中加上这么多的零作为整个多项式的根。
d = c(2:end)./c(1);  % 把c(1) 作为除数
while any(isinf(d))  % 检查结果是否有无穷大的（如果有则意味着上溢发生)
    c = c(2:end);
    d = c(2:end)./c(1);
end

% 通过多项式的友矩阵求根
n = length(c);
if n > 1
    a = zeros(n-1,n-1,class(fullPrototype)); % (n-1)x(n-1)的零矩阵
    a(1,:) = -d;   % 第一行是d 
    a(2:n:end) = 1;  
%  a 形如　
%		d(1)  d(2)  ...  d(n-1)
%	    1    0    ...   0 
%     0    1    ...   0
%     .    .          .
%     .    .          .
%     .    .          .
    r = [r;eig(a)];
end
